/*
     eib.c - Low level EIB driver for BCU 1

     Copyright (C) 2000 Bernd Thallner <bernd@kangaroo.at>

     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#define EIB_DEBUG

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>

#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/serial_reg.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/moduleparam.h>
#include <linux/wait.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include"eib.h"

// EXPORT_NO_SYMBOLS;

static char *eib_name="Low level EIB driver for BCU 1";
static char *eib_version="0.2";

#define EIB_BCU_VERSION 1   /* eib bcu version */

#define MAX_READ_BUFFER 20  /* * 33 byte buffer size */
#define MAX_WRITE_BUFFER 5  /* * 33 byte buffer size */

// maximum time in usec to wait between rx interrupt raised and byte fully send
#define MAX_TEMT_WAIT 1000

// minimum time in usec to wait between byte fully send and clear rts signal
#define WRITE_TEMT_CLRRTS_DELAY 100

// minimum time in usec to wait between clear rts and set rts signal
#define WRITE_CLRRTS_SETRTS_DELAY 80

// timeout for data block transfer
#define READ_WRITE_TIMEOUT 130000

// maximum time between to interrupts if not in idle state
#define MAX_TIME_BETWEEN_INT 80000

// minimum time to wait after reset without any respond to signals
#define MIN_RESET_WAIT_TIME 130000

// enable debug messsages
#define EIB_DEBUG

// sniff_falcon_hack
// #define SNIFF_HACK

#if !defined(DEFAULT_IO)
  #define DEFAULT_IO 0x03f8
#endif

#if !defined(DEFAULT_IRQ)
  #define DEFAULT_IRQ 4
#endif

#if !defined(DEFAULT_MAJOR)
  #define DEFAULT_MAJOR 121
#endif

MODULE_LICENSE("GPL v2");

/*
 * Function headers
 *
 */
static ssize_t eib_read(struct file *file, char *buf, size_t count, loff_t *offset);

static ssize_t eib_write(struct file *file, const char *buf, size_t count, loff_t *offset);

static unsigned int eib_poll(struct file *file, poll_table *wait);

static int eib_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);

static int eib_open(struct inode *inode, struct file *file);

static int eib_flush(struct file *file);

static int eib_release(struct inode *inode, struct file *file);

void initate_send_timer(unsigned long data);

void detect_reset_timer(unsigned long data);

void do_temt_bh(unsigned long unused);     /* tasklet scheduled in the Interrupt handler */

irqreturn_t interrupt_handler(int irq, void *dev_id, struct pt_regs *regs);




#define setRTS() outb(UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2, eib_io + UART_MCR)
#define clrRTS() outb(UART_MCR_DTR | UART_MCR_OUT2, eib_io+UART_MCR)

#define getCTS() (inb(eib_io + UART_MSR) & UART_MSR_CTS)


#define add_initate_send_timer()  { del_timer(&initate_send_timer_list); \
                                    init_timer(&initate_send_timer_list); \
                                    initate_send_timer_list.expires = jiffies + 1; \
                                    initate_send_timer_list.function = \
                                    initate_send_timer; \
                                    add_timer(&initate_send_timer_list); \
                                  }

#define add_detect_reset_timer()  { del_timer(&detect_reset_timer_list); \
                                    init_timer(&detect_reset_timer_list); \
                                    detect_reset_timer_list.expires = jiffies + 1; \
                                    detect_reset_timer_list.function= \
                                    detect_reset_timer; \
                                    add_timer(&detect_reset_timer_list); \
                                  }

/*
 * File Operation structure
 * 
 */
struct file_operations eib_fops = 
  {
    read:       eib_read,
    write:      eib_write,
    poll:       eib_poll,
    ioctl:      eib_ioctl,
    open:       eib_open,
    flush:      eib_flush,
    release:    eib_release,
  };


int eib_io    = DEFAULT_IO;
int eib_irq   = DEFAULT_IRQ;
int eib_major = DEFAULT_MAJOR;


MODULE_AUTHOR("Bernd Thallner, bernd@kangaroo.at");
MODULE_DESCRIPTION("A low level EIB driver for BCU 1.");
MODULE_PARM(eib_io, "i");
MODULE_PARM_DESC(eib_io, "I/O port for the serial port where the BCU 1 is connected.");
MODULE_PARM(eib_irq, "i");
MODULE_PARM_DESC(eib_irq, "Interrupt number for the serial port where the BCU 1 is connected.");
MODULE_PARM(eib_major, "i");
MODULE_PARM_DESC(eib_major, "Major number for this device.");


#ifdef SNIFF_HACK
  int ignore_read_forced = 0;
  MODULE_PARM(ignore_read_forced, "i");
#endif

unsigned long warning_flag   = 0;
unsigned char read_buffer [MAX_READ_BUFFER][33];
unsigned char write_buffer[MAX_WRITE_BUFFER][33];

int rb_col                   = 0;
int wb_col                   = 0;

/* State 0 : CTS = 1 / RTS = 1
 *
 *
 *
 */
volatile int state           = 0;  /* actual state of the state-machine */
volatile int rb_head         = 0;
volatile int rb_length       = 0;
volatile int wb_tail         = 0; 
volatile int wb_length       = 0; 
volatile int initate_send    = 0;
volatile int eib_read_count  = 0;
volatile int eib_write_count = 0;
volatile int eib_single_open = 0;  /* Multiple access allowed! */
volatile int eib_open_count  = 0;

struct timeval last_int_tv;
struct timeval start_tv;
struct timeval reset_tv;           /* Time of the last reset */

struct timer_list initate_send_timer_list;
struct timer_list detect_reset_timer_list;

struct eib_device_information eib_dev_info;

DECLARE_WAIT_QUEUE_HEAD(eib_read_queue);  /* wait queue */
DECLARE_WAIT_QUEUE_HEAD(eib_write_queue);


/*
 * Save uart register settings for the serial port
 */
unsigned char old_dll;    /* dll = divisor latch low       */
unsigned char old_dlh;    /* dlh = divisor latch high      */
unsigned char old_lcr;    /* lcr = line control register   */
unsigned char old_mcr;    /* mcr = modem control register  */


/*
 * Definition of a tasklet instead of a Bottom Half Interrupt handler
 */
DECLARE_TASKLET(temt_tasklet, do_temt_bh, 0);


unsigned long prcnt = 0;
void ir_print_tasklet(unsigned long);
DECLARE_TASKLET(ir_tasklet, ir_print_tasklet, 0);


spinlock_t splock = SPIN_LOCK_UNLOCKED;




/*********************************************************************
 ** Name    : bitcount
 ** Purpose : counts the 1s of a Byte 
 **
 **
 **
 ********************************************************************/
int bitcount(unsigned char x) 
{
  int b;
 
  for(b = 0; x != 0; x >>= 1)
    if(x & 01) b++;
 
  return b;
}

/*********************************************************************
 ** Name    : time_difference
 ** Purpose : if second parameter == NULL => actual time will be used
 **
 **
 **
 ********************************************************************/
inline long time_difference(struct timeval *tv1, struct timeval *tv2) 
{
  struct timeval foo;

  if(tv2 == NULL) {
    do_gettimeofday(&foo); 
    tv2 = &foo;
  }

  return     (tv2->tv_usec - tv1->tv_usec) +  /* usecs   */
    1000000 *(tv2->tv_sec - tv1->tv_sec);     /* seconds * 1000000 */
}


/*********************************************************************
 ** Name    : do_reset
 ** Purpose : Resets the serial port
 **
 **
 **
 ********************************************************************/
void do_reset(unsigned long reason) 
{

#ifdef EIB_DEBUG
  printk(KERN_NOTICE "debug: do_reset: state=%d, reason=%lx, initate_send=%d, getCTS()=%d, MCR=%x\n", 
	 state, 
	 reason, 
	 initate_send, 
	 getCTS(), 
	 inb(eib_io + UART_MCR));
#endif

  warning_flag |= reason;
  state = 7;
  inb(eib_io + UART_IIR);
  inb(eib_io + UART_RX);
  inb(eib_io + UART_MSR);
  do_gettimeofday(&reset_tv);
  clrRTS();
}

/*********************************************************************
 ** Name    : init_serial
 ** Purpose : No serial driver used/this module programs the UART
 **           itself (Speed: 9600Baud, 8 Databits, 1 Stoppbit)
 **           
 **
 ********************************************************************/
void init_serial(void) 
{
  outb(0x00, eib_io + UART_IER);
  old_lcr = inb(eib_io + UART_LCR);        /* save old line control reg */

  outb(UART_LCR_DLAB,  eib_io + UART_LCR); /* set divisor latch access bit in line control reg */
  old_dll =        inb(eib_io + UART_DLL); /* save divisor latch low */
  old_dlh =        inb(eib_io + UART_DLM); /* save divisor latch high 
					    * 115200 / Divisorlatch = speed
					    */
  outb(0x0c,           eib_io + UART_DLL); /* write 0x0c to divisor latch low */
  outb(0x00,           eib_io + UART_DLM); /* write 0x00 to divisor latch high 
					    * Speed : 9600
					    */
  
  outb(UART_LCR_WLEN8, eib_io + UART_LCR); /* 8 Databits in the line control reg */
  outb(0x00,           eib_io + UART_FCR); /* disable Fifo control reg */
  old_mcr =        inb(eib_io + UART_MCR); /* save modem control reg */

  outb(UART_MCR_DTR | UART_MCR_OUT2, eib_io + UART_MCR);  
  /* Set RTS, Enable GP02 neccessary for UART ints */
  outb(UART_IER_MSI | UART_IER_RDI,  eib_io + UART_IER);  
  /* Enable modem status int and receiver data int */

  inb(eib_io + UART_RX);  /* read and trash old byte from buffer */
}

/*********************************************************************
 ** Name    : restore_serial
 ** Purpose : Restores the settings of the UART control regs from the
 **           variables explained above
 **
 **
 ********************************************************************/
void restore_serial(void) {
  outb(0x00,          eib_io + UART_IER);
  outb(UART_LCR_DLAB, eib_io + UART_LCR); /* enable divisor latch access */
  outb(old_dll,       eib_io + UART_DLL);
  outb(old_dlh,       eib_io + UART_DLM);
  outb(old_lcr,       eib_io + UART_LCR);
  outb(old_mcr,       eib_io + UART_MCR);
}


/*********************************************************************
 ** Name    : eib_read
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static ssize_t eib_read(struct file *file, 
			char *buf, 
			size_t count, 
			loff_t *offset) 
{
  int rb_tail, len;
  unsigned long flags;
  
  while(!rb_length || eib_read_count) 
    { 
      if(file->f_flags & O_NONBLOCK) /* read buffer empty */
	return -EAGAIN;
	wait_event_interruptible(eib_read_queue, !(!rb_length || eib_read_count) );
//                 interruptible_sleep_on(&eib_read_queue);

      if(signal_pending(current))
	return -ERESTARTSYS;
    }
  
  spin_lock_irqsave(&splock, flags);
  

  if(eib_read_count) 
    {
      spin_unlock_irqrestore(&splock, flags);

      return -EAGAIN;
    }

  eib_read_count++;
  rb_tail = (rb_head-rb_length + MAX_READ_BUFFER) % MAX_READ_BUFFER;
  len = (read_buffer[rb_tail][0] & 0x1f);

  spin_unlock_irqrestore(&splock, flags);

  if(count < len) return -ENOMSG;

  if(copy_to_user(buf, &read_buffer[rb_tail][1], len))
    return -EFAULT;

  spin_lock_irqsave(&splock, flags);

  rb_length--;
  eib_read_count = 0;

  spin_unlock_irqrestore(&splock, flags);


  return len;
}


/*********************************************************************
 ** Name    : eib_write
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static ssize_t eib_write(struct file *file, 
			 const char *buf, 
			 size_t count, 
			 loff_t *offset) 
{
  int temp, wb_head;
  unsigned long flags;

  if(count > 32) return -EINVAL;
  if(rb_length >= ( MAX_READ_BUFFER - 3 ) ) return -EINVAL;
  while(wb_length == MAX_WRITE_BUFFER || 
	eib_write_count)     {             /* write buffer full */

    if(file->f_flags & O_NONBLOCK)
      return -EAGAIN;
    wait_event_interruptible(eib_write_queue,  !( (wb_length == MAX_WRITE_BUFFER) || eib_write_count));
//     interruptible_sleep_on(&eib_write_queue);
    
    if(signal_pending(current))
      return -ERESTARTSYS;
  }

  spin_lock_irqsave(&splock, flags);

  if(eib_write_count) {

    spin_unlock_irqrestore(&splock, flags);

    return -EAGAIN;
  }

  eib_write_count++;

  wb_head = (wb_tail + wb_length) % MAX_WRITE_BUFFER;

  spin_unlock_irqrestore(&splock, flags);

  temp = (count & 0x1f) | 0x20;
  write_buffer[wb_head][0] = temp | (bitcount(temp) % 2) * 0x80;
  if(copy_from_user(&write_buffer[wb_head][1], buf, count))
    return -EFAULT;

  spin_lock_irqsave(&splock, flags);

  wb_length++;
  add_initate_send_timer();
  eib_write_count = 0;

  spin_unlock_irqrestore(&splock, flags);

  return count & 0x1f;
}




/*********************************************************************
 ** Name    : eib_poll
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static unsigned int eib_poll(struct file *file, poll_table *wait) 
{
  unsigned int mask = 0;

  poll_wait(file, &eib_read_queue, wait);
  poll_wait(file, &eib_write_queue, wait);

  if(rb_length)  mask |= POLLIN | POLLRDNORM;

  if(wb_length != MAX_WRITE_BUFFER)
    mask |= POLLOUT | POLLWRNORM;

  return mask;
}




/*********************************************************************
 ** Name    : eib_ioctl
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static int eib_ioctl(struct inode *inode, 
		     struct file *file, 
		     unsigned int cmd, 
		     unsigned long arg)
{
  unsigned long flags;

  switch(cmd) {
    
  case EIB_GET_BCU_VERSION:
    return EIB_BCU_VERSION;
    
  case EIB_GET_DEVICE_INFORMATION:
    eib_dev_info.warning_flag = warning_flag;
    eib_dev_info.rb_length = rb_length;
    eib_dev_info.wb_length = wb_length;

    if(copy_to_user((void *)arg, (void *)&eib_dev_info, sizeof(struct eib_device_information)))
      return -EFAULT;

    break;

  case EIB_RESET_DEVICE_STATISTIC:
    spin_lock_irqsave(&splock, flags);

    warning_flag = 0;
    eib_dev_info.stat_rx_bytes = 0;
    eib_dev_info.stat_tx_bytes = 0;
    eib_dev_info.stat_rx_data_blocks = 0;
    eib_dev_info.stat_tx_data_blocks = 0;
    eib_dev_info.stat_max_rx_time_data_block = 0;
    eib_dev_info.stat_max_tx_time_data_block = 0;

    spin_unlock_irqrestore(&splock, flags);

    break;

  case EIB_CLEAR_WRITE_BUFFER:
    spin_lock_irqsave(&splock, flags);

    del_timer(&initate_send_timer_list);
    add_detect_reset_timer();
    initate_send=0;
    warning_flag=0;
    wb_length=0;
    do_reset(0x00400000);

    spin_unlock_irqrestore(&splock, flags);

    wake_up_interruptible(&eib_write_queue);
    break;

  case EIB_CLEAR_READ_BUFFER:
    spin_lock_irqsave(&splock, flags);

    del_timer(&initate_send_timer_list);
    add_detect_reset_timer();
    warning_flag=0;
    rb_length=0;
    do_reset(0x00400000);

    spin_unlock_irqrestore(&splock, flags);

    wake_up_interruptible(&eib_read_queue);
    break;

  case EIB_SINGLE_OPEN:
    if(eib_open_count>1) return eib_open_count;
    eib_single_open=1;

    break;

  default:
    return -ENOIOCTLCMD;
  }

  return 0;
}




/*********************************************************************
 ** Name    : eib_open
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static int eib_open(struct inode *inode, struct file *file) 
{
  if(eib_single_open) return -EBUSY;
//   MOD_INC_USE_COUNT; 
  eib_open_count++;
  return 0;
}



/*********************************************************************
 ** Name    : eib_flush
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static int eib_flush(struct file *file) 
{  
  while(wb_length) { // write buffer not empty
    wait_event_interruptible(eib_write_queue, 1==1);   //   interruptible_sleep_on(&eib_write_queue);
    if(signal_pending(current))
      return -ERESTARTSYS;
  }

  return 0;
}



/*********************************************************************
 ** Name    : eib_release
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static int eib_release(struct inode *inode, struct file *file) 
{
  eib_open_count--;
  eib_single_open = 0;
//   MOD_DEC_USE_COUNT; 
  return 0;

}

/*********************************************************************
 ** Name    : eib_init_module
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
int __init eib_init_module(void) 
{
  int retval;

//   if((retval = check_region(eib_io, 0x08)) < 0)
//     return retval;
  
  if(request_region(eib_io, 0x08, "eib") < 0) return retval;;   /* Serial Conn. needs 8 Byte */
//   request_region(eib_io, 0x08, "eib");   /* Serial Conn. needs 8 Byte */
  outb(0x00, eib_io+UART_IER);
   
  retval = request_irq(eib_irq, interrupt_handler, SA_INTERRUPT, "eib", &eib_fops);

//   retval = request_irq(eib_irq, interrupt_handler, SA_INTERRUPT, "eib", NULL);

  if(retval) {
    release_region(eib_io, 0x08);
    return retval;
  }

  retval = register_chrdev(eib_major, "eib", &eib_fops);
  if(retval<0) {
    release_region(eib_io, 0x08);
    free_irq(eib_irq, &eib_fops);
//     free_irq(eib_irq, NULL);
    return retval;
  }

  if(eib_major == 0)  eib_major = retval;

  init_serial();

  memset(&eib_dev_info, 0, sizeof(struct eib_device_information));
  sprintf(eib_dev_info.string, "%s, version=%s, io=0x%x, irq=%d, major=%d", 
	  eib_name, 
	  eib_version, 
	  eib_io, 
	  eib_irq, 
	  eib_major);


  printk(KERN_INFO "%s\n", eib_dev_info.string);

  eib_dev_info.rb_max_length = MAX_READ_BUFFER;
  eib_dev_info.wb_max_length = MAX_WRITE_BUFFER;
  eib_dev_info.io = eib_io;
  eib_dev_info.irq = eib_irq;
  eib_dev_info.major = eib_major;

  do_gettimeofday(&last_int_tv);
  init_timer(&initate_send_timer_list);
  init_timer(&detect_reset_timer_list);

  return 0;
}


/*********************************************************************
 ** Name    : eib_cleanup_module
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
void __exit eib_cleanup_module(void) 
{

#ifdef EIB_DEBUG
  printk(KERN_DEBUG "debug: cleanup_module: state=%d, warning_flag=%lx, initate_send=%d, getCTS()=%d, MCR=%x\n", 
	 state, 
	 warning_flag, 
	 initate_send, 
	 getCTS(), 
	 inb(eib_io + UART_MCR));
#endif

  if(rb_length)
    printk(KERN_WARNING "eib: warning: read buffer was not empty.\n");

  del_timer(&initate_send_timer_list);
  del_timer(&detect_reset_timer_list);
  restore_serial();
  release_region(eib_io, 0x08);
  free_irq(eib_irq, &eib_fops);
//   free_irq(eib_irq, NULL);
  unregister_chrdev(eib_major, "eib");

  printk(KERN_INFO "eib successfully unloaded!\n");
}




/*********************************************************************
 ** Name    : initate_send_timer
 ** Purpose : ?
 **
 **
 **
 ********************************************************************/
void initate_send_timer(unsigned long data) 
{
  unsigned long flags;

  if(wb_length == 0) {
    warning_flag |= 0x10000; // initate send with empty write buffer
    return;
  }

  spin_lock_irqsave(&splock, flags);

  if(state==0 && !getCTS() && 
     time_difference(&last_int_tv, NULL) > 3000) {

    initate_send=1;
    do_gettimeofday(&start_tv);
    do_gettimeofday(&last_int_tv);

    setRTS();
    add_detect_reset_timer();

    spin_unlock_irqrestore(&splock, flags);

    return;
  }
  add_initate_send_timer();

  spin_unlock_irqrestore(&splock, flags);
}




/*********************************************************************
 ** Name    : detect_reset_timer
 ** Purpose : ?
 **
 **
 **
 ********************************************************************/
void detect_reset_timer(unsigned long data) 
{
  unsigned long flags;

  spin_lock_irqsave(&splock, flags);

  if(state==0 && !initate_send && !getCTS()) { // idle

    spin_unlock_irqrestore(&splock, flags);

    return;
  } else if(state==7) { // reset
    if(time_difference(&reset_tv, NULL)>MIN_RESET_WAIT_TIME) {
      do_gettimeofday(&last_int_tv);
      if(wb_length)
	add_initate_send_timer();
      rb_col=0;
      wb_col=0;
      state=0;

      spin_unlock_irqrestore(&splock, flags);

      return;
    }
  } else { // read || write
    if(time_difference(&start_tv, NULL)>READ_WRITE_TIMEOUT)
      do_reset(0x100000); // read write timeout
    else if(time_difference(&last_int_tv, NULL)>MAX_TIME_BETWEEN_INT)
      do_reset(0x200000); // max time between int exceeded
  }
  add_detect_reset_timer();

  spin_unlock_irqrestore(&splock, flags);
}




/*********************************************************************
 ** Name    : check_interrupt_pending
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
inline void check_interrupt_pending(void) 
{
  int id = inb(eib_io + UART_IIR);

  if(~id & UART_IIR_NO_INT) {    /* = 0x01 no int occured */
    if((id & UART_IIR_ID) == UART_IIR_RDI)
      do_reset(0x40000); // unexpected receive data interrupt

    else 
      if(inb(eib_io + UART_MSR) & UART_MSR_DCTS)
	if(getCTS())
	  do_reset(0x80000); // unexpected cts change to high
      // else !getCTS() -> state unchanged
    // else uninteresting modem status register change -> ignore
  }
}



/*********************************************************************
 ** Name    : do_temt_bh
 ** Purpose : Tasklet Bottom half interrupt handler
 **
 **
 **
 ********************************************************************/
void do_temt_bh(unsigned long unused) 
{
  struct timeval tv;
  unsigned long flags;
  long temp;

  if(state == 7) return;        /* reset */

  if(state != 3 && state != 6) {
    do_reset(0x40);             /* unknown state */
    return;
  }

  do_gettimeofday(&tv);
  while((~inb(eib_io+UART_LSR) & UART_LSR_TEMT) && 
	time_difference(&tv, NULL)<MAX_TEMT_WAIT)
    ;                           /* busy waiting */

  if(~inb(eib_io+UART_LSR)&UART_LSR_TEMT)
    do_reset(0x1000); // exceed MAX_TEMT_WAIT

  switch(state) {

  case 3:
    eib_dev_info.stat_rx_bytes++;
    if(rb_col==(read_buffer[rb_head][0]&0x1f)+1) { // read finished
      eib_dev_info.stat_rx_data_blocks++;
      if(eib_dev_info.stat_max_rx_time_data_block<(temp=time_difference(&start_tv, NULL)))
	eib_dev_info.stat_max_rx_time_data_block=temp;

      if(read_buffer[rb_head][0]==0xa0)
	warning_flag |= 0x20000; // read bcu reset
      else 
	{
	  rb_head = (rb_head + 1) % MAX_READ_BUFFER;
	  rb_length++;
	  wake_up_interruptible(&eib_read_queue);
	}

      rb_col=0;

      spin_lock_irqsave(&splock, flags);

      if(wb_length)
	add_initate_send_timer();
      state = 0;
    } 
    else 
      {
	spin_lock_irqsave(&splock, flags);
	state = 1;
      }
    check_interrupt_pending();
    clrRTS();

    spin_unlock_irqrestore(&splock, flags);

    break;

  case 6:
    do_gettimeofday(&tv);

    while(time_difference(&tv, NULL) < WRITE_TEMT_CLRRTS_DELAY);

    eib_dev_info.stat_tx_bytes++;

    if(wb_col==(write_buffer[wb_tail][0]&0x1f)+1) { // write finished
      eib_dev_info.stat_tx_data_blocks++;
      if(eib_dev_info.stat_max_tx_time_data_block<(temp=time_difference(&start_tv, NULL)))
	eib_dev_info.stat_max_tx_time_data_block=temp;

      wb_tail=(wb_tail+1)%MAX_WRITE_BUFFER;
      wb_length--;
      wake_up_interruptible(&eib_write_queue);
      wb_col=0;

      if(wb_length)
	add_initate_send_timer();

      spin_lock_irqsave(&splock, flags);


      state = 0;

      check_interrupt_pending();
      clrRTS();

    } /* if(wb_col... */
    else {

      check_interrupt_pending();

      clrRTS();
      do_gettimeofday(&tv);
      while(time_difference(&tv, NULL) < WRITE_CLRRTS_SETRTS_DELAY);

      spin_lock_irqsave(&splock, flags);

      state=4;

      check_interrupt_pending();
      setRTS();
    }

    spin_unlock_irqrestore(&splock, flags);

    break;

  case 7: // reset
    break;

  default:
    do_reset(0x40); // unknown state
    break;
  }
  do_gettimeofday(&last_int_tv);
}



/*********************************************************************
 ** Name    : ir_print_tasklet
 ** Purpose : ?
 **
 **
 **
 ********************************************************************/
void ir_print_tasklet(unsigned long unused)
{
  printk(KERN_INFO "Hallo! + %ld\n", prcnt);
  
  ++prcnt;

}



/*********************************************************************
 ** Name    : interrupt_handler
 ** Purpose : Performs the state machine
 **
 **
 **
 ********************************************************************/
irqreturn_t interrupt_handler(int irq, 
		       void *dev_id, 
		       struct pt_regs *regs) 
{
  int tmp = 0, id;

  id = inb(eib_io + UART_IIR);

  if(~id & UART_IIR_NO_INT) {              /* is there any interrupt ? */

    if(state!=0 && state!=7 && 
       time_difference(&start_tv, NULL) > READ_WRITE_TIMEOUT)
      do_reset(0x100000);                  /* read write timeout */

    switch(state) {                        /* State machine state */ 
      
    case 0: // idle
      if((id & UART_IIR_ID) == UART_IIR_MSI){/* Modem status interrupt */
	if(inb(eib_io + UART_MSR) & UART_MSR_DCTS) {
	  if(getCTS()) {
	    if(initate_send) {             /* write to bcu */
	      outb(write_buffer[wb_tail][0], eib_io + UART_TX);
	      initate_send = 0;
	      wb_col++;
	      state = 5;
	      
	      break;
	    } 
	    else { // read
	      add_detect_reset_timer();
	      if(rb_length == MAX_READ_BUFFER) {
		do_reset(0x10);            /* read buffer overflow */
	      } 
	      else {
		do_gettimeofday(&start_tv);
		setRTS();
		outb(0xFF, eib_io + UART_TX);
		state = 2;
	      }
	    }
	  } // else !getCTS() -> state=0
	} // else uninteresting modem status register change -> ignore
      } 
      else do_reset(0x2);                  /* unexpected receive data interrupt */
      
      break;
      
    case 1:
      if((id & UART_IIR_ID) == UART_IIR_MSI) {
	if(inb(eib_io + UART_MSR) & UART_MSR_DCTS)
	  if(getCTS()) {
	    setRTS();
	    outb(0x00, eib_io + UART_TX);
	    state=2;
	  } // else !getCTS() -> state=1
	// else uninteresting modem status register change -> ignore
      } else
	do_reset(0x200); // unexpected receive data interrupt
      break;
      
    case 2:
      if((id & UART_IIR_ID) == UART_IIR_RDI) {
	
	tmp = inb(eib_io + UART_RX);
	read_buffer[rb_head][rb_col++] = tmp;
	if(inb(eib_io + UART_LSR) & UART_LSR_OE) {
	  do_reset(0x80); // overrun
	  break;
	}
	if(rb_col==1 && !(((tmp & 0x60) == 0x20) && 
                          ((tmp & 0x80) == (bitcount(tmp & 0x7f) % 2) * 0x80))) {
	  do_reset(0x400); // parity error
	  break;
	}
	state=3;

	tasklet_schedule(&temt_tasklet);   /* Instead of the Bottom half handler */
      }
      else if(inb(eib_io+UART_MSR)&UART_MSR_DCTS)
	if(getCTS())
	  do_reset(0x20);                  /* unexpected cts change to high */
      // else !getCTS() -> ok
      // else uninteresting modem status register change
      break;
      
    case 3:

      if((id & UART_IIR_ID) == UART_IIR_RDI)
	do_reset(0x4);                     /* unexpected receive data interrupt */
      
      else 
	if(inb(eib_io + UART_MSR) & UART_MSR_DCTS)
	  if(getCTS()) 
	    {
	      //state = 1;
	      do_reset(0x8);                 /* unexpected cts change to high */
	    }

      break;

    case 6:

      if((id & UART_IIR_ID) == UART_IIR_RDI)
	do_reset(0x4);                     /* unexpected receive data interrupt */
      
      else 
	if(inb(eib_io + UART_MSR) & UART_MSR_DCTS)
	  if(getCTS())
	    do_reset(0x8);                 /* unexpected cts change to high */
      // else !getCTS() -> state unchanged
      // else uninteresting modem status register change -> ignore
      break;

    case 4:

      if((id&UART_IIR_ID)==UART_IIR_MSI) {
	if(inb(eib_io+UART_MSR)&UART_MSR_DCTS) 
	  if(getCTS()) {
	    outb_p(write_buffer[wb_tail][wb_col++], eib_io+UART_TX);
	    state=5;
	  } // else !getCTS() -> state=4
	// else uninteresting modem status register change -> ignore
      } else
	do_reset(0x800);                   /* unexpected receive data interrupt */
      break;
      
    case 5:
      if((id&UART_IIR_ID)==UART_IIR_RDI) {
	tmp = inb(eib_io+UART_RX);

	state=6;

	if(inb(eib_io+UART_LSR)&UART_LSR_OE) {
	  do_reset(0x80);                  /* overun */
	  break;
	}
	
#ifdef SNIFF_HACK
	if(wb_col==1 && tmp!=0xff && !ignore_read_forced) {
#else
	if(wb_col==1 && tmp!=0xff) {
#endif
	    
	  warning_flag|=0x100;             /* read forced */
	  wb_col=0;
	  if(rb_length==MAX_READ_BUFFER) {
	    do_reset(0x4000);              /* read forced && read buffer full */
	    
	    break;
	  }
	  
	  if(((tmp & 0x60) != 0x20) ||
	     ((tmp & 0x80) != (bitcount(tmp & 0x7f) % 2) * 0x80)) {
	    do_reset(0x400);               /* parity error */
	    
	    break;
	  }
	  read_buffer[rb_head][rb_col++] = tmp;

	  state=3;

	}
	if(wb_col>1 && tmp!=0x00) {
	  do_reset(0x2000);                /* unexpect byte received */
	  break;
	}

	tasklet_schedule(&temt_tasklet);   /* Instead of the Bottom half handler */
      } 
      else 
	if(inb(eib_io+UART_MSR)&UART_MSR_DCTS)
	  do_reset(0x8000);                /* unexpected cts change */
	// else uninteresting modem status register change
	break;
	
    case 7: // reset
      if((id&UART_IIR_ID)==UART_IIR_MSI)
	inb(eib_io + UART_MSR);
      else inb(eib_io + UART_RX);

      break;
      
      default:
	do_reset(0x40); // unknown state	
    }
  } 
  else warning_flag |= 0x1; // no interrupt pending

    //    tasklet_schedule(&ir_tasklet);


  do_gettimeofday(&last_int_tv);
  return IRQ_HANDLED;
}
  


module_init(eib_init_module);   /* own function eib_init_module instead of
				 * "init_module"
				 */
module_exit(eib_cleanup_module);
